@@ -41,14 +41,17 @@ static GEN
 gen_RgXQ_eval_powers(GEN P, GEN V, long a, long n, void *E, const struct bb_algebra *ff,
                                            GEN cmul(void *E, GEN P, long a, GEN x))
 {
-  pari_sp av = avma, lim=stack_lim(av,2);
+  pari_sp av = avma;
   long i;
   GEN z = cmul(E,P,a,ff->one(E));
+  if (!z) z = gen_0;
   for (i=1; i<=n; i++)
   {
-    z = ff->add(E, z, cmul(E,P,a+i,gel(V,i+1)));
-    if (low_stack(lim, stack_lim(av,2)))
-      z = gerepileupto(av, z);
+    GEN t = cmul(E,P,a+i,gel(V,i+1));
+    if (t) {
+      z = ff->add(E, z, t);
+      if (gc_needed(av,2)) z = gerepileupto(av, z);
+    }
   }
   return ff->red(E,z);
 }
@@ -62,7 +65,7 @@ GEN
 gen_bkeval_powers(GEN P, long d, GEN V, void *E, const struct bb_algebra *ff,
                                      GEN cmul(void *E, GEN P, long a, GEN x))
 {
-  pari_sp av = avma, lim;
+  pari_sp av = avma;
   long l = lg(V)-1;
   GEN z, u;
 
@@ -71,13 +74,12 @@ gen_bkeval_powers(GEN P, long d, GEN V, void *E, const struct bb_algebra *ff,
   if (l<2) pari_err_DOMAIN("gen_RgX_bkeval_powers", "#powers", "<",gen_2,V);
   d -= l;
   z = gen_RgXQ_eval_powers(P,V,d+1,l-1,E,ff,cmul);
-  lim = stack_lim(av,2);
   while (d >= l-1)
   {
     d -= l-1;
     u = gen_RgXQ_eval_powers(P,V,d+1,l-2,E,ff,cmul);
     z = ff->add(E,u, ff->mul(E,z,gel(V,l)));
-    if (low_stack(lim, stack_lim(av,2)))
+    if (gc_needed(av,2))
       z = gerepileupto(av, z);
   }
   u = gen_RgXQ_eval_powers(P,V,0,d,E,ff,cmul);
@@ -104,6 +106,59 @@ gen_bkeval(GEN Q, long d, GEN x, int use_sqr, void *E, const struct bb_algebra *
   return gerepileupto(av, z);
 }
 
+static GEN
+_gen_nored(void *E, GEN x) { (void)E; return x; }
+static GEN
+_gen_add(void *E, GEN x, GEN y) { (void)E; return gadd(x, y); }
+static GEN
+_gen_sub(void *E, GEN x, GEN y) { (void)E; return gsub(x, y); }
+static GEN
+_gen_mul(void *E, GEN x, GEN y) { (void)E; return gmul(x, y); }
+static GEN
+_gen_sqr(void *E, GEN x) { (void)E; return gsqr(x); }
+static GEN
+_gen_one(void *E) { (void)E; return gen_1; }
+static GEN
+_gen_zero(void *E) { (void)E; return gen_0; }
+
+static struct bb_algebra Rg_algebra = { _gen_nored, _gen_add, _gen_sub,
+              _gen_mul, _gen_sqr,_gen_one,_gen_zero };
+
+static GEN
+_gen_cmul(void *E, GEN P, long a, GEN x)
+{(void)E; return gmul(gel(P,a+2), x);}
+
+GEN
+RgX_RgV_eval(GEN Q, GEN x)
+{
+  return gen_bkeval_powers(Q, degpol(Q), x, NULL, &Rg_algebra, _gen_cmul);
+}
+
+GEN
+RgX_Rg_eval_bk(GEN Q, GEN x)
+{
+  return gen_bkeval(Q, degpol(Q), x, 1, NULL, &Rg_algebra, _gen_cmul);
+}
+
+GEN
+RgXV_RgV_eval(GEN Q, GEN x)
+{
+  long i, l = lg(Q), vQ = gvar(Q);
+  GEN v = cgetg(l, t_VEC);
+  for (i = 1; i < l; i++)
+  {
+    GEN Qi = gel(Q, i);
+    gel(v, i) = typ(Qi)==t_POL && varn(Qi)==vQ? RgX_RgV_eval(Qi, x): gcopy(Qi);
+  }
+  return v;
+}
+
+const struct bb_algebra *
+get_Rg_algebra(void)
+{
+  return &Rg_algebra;
+}
+
 /*******************************************************************/
 /*                                                                 */
 /*                         RgX                                     */
@@ -216,6 +271,24 @@ ZX_unscale(GEN P, GEN h)
   }
   return Q;
 }
+/* P a ZX. Return P(x << n), not memory clean */
+GEN
+ZX_unscale2n(GEN P, long n)
+{
+  long i, ni = n, l = lg(P);
+  GEN Q = cgetg(l, t_POL);
+  Q[1] = P[1];
+  if (l == 2) return Q;
+  gel(Q,2) = gel(P,2);
+  if (l == 3) return Q;
+  gel(Q,3) = shifti(gel(P,3), ni);
+  for (i=4; i<l; i++)
+  {
+    ni += n;
+    gel(Q,i) = shifti(gel(P,i), ni);
+  }
+  return Q;
+}
 /* P(h*X) / h, assuming h | P(0), i.e. the result is a ZX */
 GEN
 ZX_unscale_div(GEN P, GEN h)
@@ -299,20 +372,19 @@ RgX_inflate(GEN x0, long d)
 GEN
 RgX_translate(GEN P, GEN c)
 {
-  pari_sp av = avma, lim;
+  pari_sp av = avma;
   GEN Q, *R;
   long i, k, n;
 
   if (!signe(P) || gequal0(c)) return RgX_copy(P);
   Q = leafcopy(P);
   R = (GEN*)(Q+2); n = degpol(P);
-  lim = stack_lim(av, 2);
   if (gequal1(c))
   {
     for (i=1; i<=n; i++)
     {
       for (k=n-i; k<n; k++) R[k] = gadd(R[k], R[k+1]);
-      if (low_stack(lim, stack_lim(av,2)))
+      if (gc_needed(av,2))
       {
         if(DEBUGMEM>1) pari_warn(warnmem,"TR_POL(1), i = %ld/%ld", i,n);
         Q = gerepilecopy(av, Q); R = (GEN*)Q+2;
@@ -324,7 +396,7 @@ RgX_translate(GEN P, GEN c)
     for (i=1; i<=n; i++)
     {
       for (k=n-i; k<n; k++) R[k] = gsub(R[k], R[k+1]);
-      if (low_stack(lim, stack_lim(av,2)))
+      if (gc_needed(av,2))
       {
         if(DEBUGMEM>1) pari_warn(warnmem,"TR_POL(-1), i = %ld/%ld", i,n);
         Q = gerepilecopy(av, Q); R = (GEN*)Q+2;
@@ -336,7 +408,7 @@ RgX_translate(GEN P, GEN c)
     for (i=1; i<=n; i++)
     {
       for (k=n-i; k<n; k++) R[k] = gadd(R[k], gmul(c, R[k+1]));
-      if (low_stack(lim, stack_lim(av,2)))
+      if (gc_needed(av,2))
       {
         if(DEBUGMEM>1) pari_warn(warnmem,"TR_POL, i = %ld/%ld", i,n);
         Q = gerepilecopy(av, Q); R = (GEN*)Q+2;
@@ -350,20 +422,19 @@ RgX_translate(GEN P, GEN c)
 GEN
 ZX_translate(GEN P, GEN c)
 {
-  pari_sp av = avma, lim;
+  pari_sp av = avma;
   GEN Q, *R;
   long i, k, n;
 
   if (!signe(P) || !signe(c)) return ZX_copy(P);
   Q = leafcopy(P);
   R = (GEN*)(Q+2); n = degpol(P);
-  lim = stack_lim(av, 2);
   if (equali1(c))
   {
     for (i=1; i<=n; i++)
     {
       for (k=n-i; k<n; k++) R[k] = addii(R[k], R[k+1]);
-      if (low_stack(lim, stack_lim(av,2)))
+      if (gc_needed(av,2))
       {
         if(DEBUGMEM>1) pari_warn(warnmem,"ZX_translate(1), i = %ld/%ld", i,n);
         Q = gerepilecopy(av, Q); R = (GEN*)Q+2;
@@ -375,7 +446,7 @@ ZX_translate(GEN P, GEN c)
     for (i=1; i<=n; i++)
     {
       for (k=n-i; k<n; k++) R[k] = subii(R[k], R[k+1]);
-      if (low_stack(lim, stack_lim(av,2)))
+      if (gc_needed(av,2))
       {
         if(DEBUGMEM>1) pari_warn(warnmem,"ZX_translate(-1), i = %ld/%ld", i,n);
         Q = gerepilecopy(av, Q); R = (GEN*)Q+2;
@@ -387,7 +458,7 @@ ZX_translate(GEN P, GEN c)
     for (i=1; i<=n; i++)
     {
       for (k=n-i; k<n; k++) R[k] = addii(R[k], mulii(c, R[k+1]));
-      if (low_stack(lim, stack_lim(av,2)))
+      if (gc_needed(av,2))
       {
         if(DEBUGMEM>1) pari_warn(warnmem,"ZX_translate, i = %ld/%ld", i,n);
         Q = gerepilecopy(av, Q); R = (GEN*)Q+2;
@@ -400,14 +471,13 @@ ZX_translate(GEN P, GEN c)
 GEN
 RgXQX_translate(GEN P, GEN c, GEN T)
 {
-  pari_sp av = avma, lim;
+  pari_sp av = avma;
   GEN Q, *R;
   long i, k, n;
 
   if (!signe(P) || gequal0(c)) return RgX_copy(P);
   Q = leafcopy(P);
   R = (GEN*)(Q+2); n = degpol(P);
-  lim = stack_lim(av, 2);
   for (i=1; i<=n; i++)
   {
     for (k=n-i; k<n; k++)
@@ -415,7 +485,7 @@ RgXQX_translate(GEN P, GEN c, GEN T)
       pari_sp av2 = avma;
       R[k] = gerepileupto(av2, RgX_rem(gadd(R[k], gmul(c, R[k+1])), T));
     }
-    if (low_stack(lim, stack_lim(av,2)))
+    if (gc_needed(av,2))
     {
       if(DEBUGMEM>1) pari_warn(warnmem,"RgXQX_translate, i = %ld/%ld", i,n);
       Q = gerepilecopy(av, Q); R = (GEN*)Q+2;
@@ -548,7 +618,7 @@ RgV_to_RgX_reverse(GEN x, long v)
 
 /* return the (N-dimensional) vector of coeffs of p */
 GEN
-RgX_to_RgV(GEN x, long N)
+RgX_to_RgC(GEN x, long N)
 {
   long i, l;
   GEN z;
@@ -560,9 +630,9 @@ RgX_to_RgV(GEN x, long N)
   return z;
 }
 GEN
-Rg_to_RgV(GEN x, long N)
+Rg_to_RgC(GEN x, long N)
 {
-  return (typ(x) == t_POL)? RgX_to_RgV(x,N): scalarcol_shallow(x, N);
+  return (typ(x) == t_POL)? RgX_to_RgC(x,N): scalarcol_shallow(x, N);
 }
 
 /* vector of polynomials (in v) whose coeffs are given by the columns of x */
@@ -582,7 +652,7 @@ RgV_to_RgM(GEN v, long n)
 {
   long j, N = lg(v);
   GEN y = cgetg(N, t_MAT);
-  for (j=1; j<N; j++) gel(y,j) = Rg_to_RgV(gel(v,j), n);
+  for (j=1; j<N; j++) gel(y,j) = Rg_to_RgC(gel(v,j), n);
   return y;
 }
 GEN
@@ -590,7 +660,7 @@ RgXV_to_RgM(GEN v, long n)
 {
   long j, N = lg(v);
   GEN y = cgetg(N, t_MAT);
-  for (j=1; j<N; j++) gel(y,j) = RgX_to_RgV(gel(v,j), n);
+  for (j=1; j<N; j++) gel(y,j) = RgX_to_RgC(gel(v,j), n);
   return y;
 }
 
@@ -613,7 +683,7 @@ RgXX_to_RgM(GEN v, long n)
 {
   long j, N = lg(v)-1;
   GEN y = cgetg(N, t_MAT);
-  for (j=1; j<N; j++) gel(y,j) = Rg_to_RgV(gel(v,j+1), n);
+  for (j=1; j<N; j++) gel(y,j) = Rg_to_RgC(gel(v,j+1), n);
   return y;
 }
 
@@ -650,18 +720,40 @@ RgXY_swap(GEN x, long n, long w)
   setvarn(z, varn(x)); return z;
 }
 
+long
+RgXY_degreex(GEN b)
+{
+  long deg = -1, i;
+  if (!signe(b)) return -1;
+  for (i = 2; i < lg(b); ++i)
+  {
+    GEN bi = gel(b, i);
+    if (typ(bi) == t_POL)
+      deg = maxss(deg, degpol(bi));
+  }
+  return deg;
+}
+
 /* return (x % X^n). Shallow */
 GEN
-RgX_modXn_shallow(GEN a, long n)
+RgXn_red_shallow(GEN a, long n)
 {
   long i, L, l = lg(a);
   GEN  b;
   if (l == 2 || !n) return pol_0(varn(a));
-  if (n < 0) pari_err_DOMAIN("RgX_modXn", "n", "<", gen_0, stoi(n));
   L = n+2; if (L > l) L = l;
   b = cgetg(L, t_POL); b[1] = a[1];
   for (i=2; i<L; i++) gel(b,i) = gel(a,i);
-  return b;
+  return normalizepol_lg(b,L);
+}
+
+GEN
+RgXnV_red_shallow(GEN P, long n)
+{
+  long i, l = lg(P);
+  GEN Q = cgetg(l, t_VEC);
+  for (i=1; i<l; i++) gel(Q,i) = RgXn_red_shallow(gel(P,i), n);
+  return Q;
 }
 
 /* return (x * X^n). Shallow */
@@ -736,8 +828,7 @@ RgX_mulXn(GEN x, long d)
   v = RgX_val(x);
   if (v >= d) return RgX_shift(x, -d);
   av = avma;
-  z = gred_rfrac_simple( RgX_shift_shallow(x, -v),
-                         monomial(gen_1, d - v, varn(x)));
+  z = gred_rfrac_simple(RgX_shift_shallow(x, -v), pol_xn(d - v, varn(x)));
   return gerepileupto(av, z);
 }
 
@@ -1130,7 +1221,7 @@ addmulXncopy(GEN x, GEN y, long d)
 
 /* return x * y mod t^n */
 static GEN
-RgX_mullow_basecase(GEN x, GEN y, long n)
+RgXn_mul_basecase(GEN x, GEN y, long n)
 {
   long i, lz = n+2, lx = lgpol(x), ly = lgpol(y);
   GEN z;
@@ -1151,19 +1242,19 @@ RgX_mullow_basecase(GEN x, GEN y, long n)
 }
 /* Mulders / Karatsuba product f*g mod t^n (Hanrot-Zimmermann variant) */
 GEN
-RgX_mullow(GEN f, GEN g, long n)
+RgXn_mul(GEN f, GEN g, long n)
 {
   pari_sp av = avma;
   GEN fe,fo, ge,go, l,h,m;
   long n0, n1;
   if (degpol(f) + degpol(g) < n) return RgX_mul(f,g);
-  if (n < 80) return RgX_mullow_basecase(f,g,n);
+  if (n < 80) return RgXn_mul_basecase(f,g,n);
   n0 = n>>1; n1 = n-n0;
   RgX_even_odd(f, &fe, &fo);
   RgX_even_odd(g, &ge, &go);
-  l = RgX_mullow(fe,ge,n1);
-  h = RgX_mullow(fo,go,n0);
-  m = RgX_sub(RgX_mullow(RgX_add(fe,fo),RgX_add(ge,go),n0), RgX_add(l,h));
+  l = RgXn_mul(fe,ge,n1);
+  h = RgXn_mul(fo,go,n0);
+  m = RgX_sub(RgXn_mul(RgX_add(fe,fo),RgX_add(ge,go),n0), RgX_add(l,h));
   /* n1-1 <= n0 <= n1, deg l,m <= n1-1, deg h <= n0-1
    * result is t^2 h(t^2) + t m(t^2) + l(t^2) */
   l = RgX_inflate(l,2); /* deg l <= 2n1 - 2 <= n-1 */
@@ -1269,7 +1360,7 @@ RgX_sqrspec_basecase(GEN x, long nx, long v)
 }
 /* return x^2 mod t^n */
 static GEN
-RgX_sqrlow_basecase(GEN x, long n)
+RgXn_sqr_basecase(GEN x, long n)
 {
   long i, lz = n+2, lx = lgpol(x);
   GEN z;
@@ -1285,18 +1376,18 @@ RgX_sqrlow_basecase(GEN x, long n)
 }
 /* Mulders / Karatsuba product f^2 mod t^n (Hanrot-Zimmermann variant) */
 GEN
-RgX_sqrlow(GEN f, long n)
+RgXn_sqr(GEN f, long n)
 {
   pari_sp av = avma;
   GEN fe,fo, l,h,m;
   long n0, n1;
   if (2*degpol(f) < n) return RgX_sqr(f);
-  if (n < 80) return RgX_sqrlow_basecase(f,n);
+  if (n < 80) return RgXn_sqr_basecase(f,n);
   n0 = n>>1; n1 = n-n0;
   RgX_even_odd(f, &fe, &fo);
-  l = RgX_sqrlow(fe,n1);
-  h = RgX_sqrlow(fo,n0);
-  m = RgX_sub(RgX_sqrlow(RgX_add(fe,fo),n0), RgX_add(l,h));
+  l = RgXn_sqr(fe,n1);
+  h = RgXn_sqr(fo,n0);
+  m = RgX_sub(RgXn_sqr(RgX_add(fe,fo),n0), RgX_add(l,h));
   /* n1-1 <= n0 <= n1, deg l,m <= n1-1, deg h <= n0-1
    * result is t^2 h(t^2) + t m(t^2) + l(t^2) */
   l = RgX_inflate(l,2); /* deg l <= 2n1 - 2 <= n-1 */
@@ -1383,6 +1474,20 @@ RgX_Rg_div(GEN x, GEN y) {
   return normalizepol_lg(z, lx);
 }
 GEN
+RgX_normalize(GEN x)
+{
+  GEN d = NULL;
+  long i, n = lg(x)-1;
+  for (i = n; i > 1; i--)
+  {
+    d = gel(x,i);
+    if (!gequal0(d)) break;
+  }
+  if (i == 1) return pol_0(varn(x));
+  if (i == n && isint1(d)) return x;
+  return normalizepol_lg(RgX_Rg_div(x, d), i+1);
+}
+GEN
 RgX_divs(GEN x, long y) {
   long i, lx;
   GEN z = cgetg_copy(x, &lx); z[1] = x[1];
@@ -1609,7 +1714,7 @@ RgXQX_divrem(GEN x, GEN y, GEN T, GEN *pr)
     }
     return pol_0(vx);
   }
-  lead = leading_term(y);
+  lead = leading_coeff(y);
   if (!dy) /* y is constant */
   {
     if (pr && pr != ONLY_DIVIDES)
@@ -1689,7 +1794,7 @@ int
 ZXQX_dvd(GEN x, GEN y, GEN T)
 {
   long dx, dy, dz, i, p, T_ismonic;
-  pari_sp av = avma, av2, lim;
+  pari_sp av = avma, av2;
   GEN y_lead;
 
   if (!signe(y)) pari_err_INV("ZXQX_dvd",y);
@@ -1697,7 +1802,7 @@ ZXQX_dvd(GEN x, GEN y, GEN T)
   if (typ(y_lead) == t_POL) y_lead = gel(y_lead, 2); /* t_INT */
   /* if monic, no point in using pseudo-division */
   if (gequal1(y_lead)) return signe(RgXQX_rem(x, y, T)) == 0;
-  T_ismonic = gequal1(leading_term(T));
+  T_ismonic = gequal1(leading_coeff(T));
   dx = degpol(x);
   if (dx < dy) return !signe(x);
   (void)new_chunk(2);
@@ -1707,7 +1812,7 @@ ZXQX_dvd(GEN x, GEN y, GEN T)
   for (i = 1; i <= dy; i++)
     if (!signe(gel(y,i))) gel(y,i) = NULL;
   dz = dx-dy; p = dz+1;
-  av2 = avma; lim = stack_lim(av2,1);
+  av2 = avma;
   for (;;)
   {
     GEN m, x0 = gel(x,0), y0 = y_lead, cx = content(x0);
@@ -1734,7 +1839,7 @@ ZXQX_dvd(GEN x, GEN y, GEN T)
     }
     do { x++; dx--; } while (dx >= 0 && !signe(gel(x,0)));
     if (dx < dy) break;
-    if (low_stack(lim,stack_lim(av2,1)))
+    if (gc_needed(av2,1))
     {
       if(DEBUGMEM>1) pari_warn(warnmem,"ZXQX_dvd dx = %ld >= %ld",dx,dy);
       gerepilecoeffs(av2,x,dx+1);
@@ -1748,7 +1853,7 @@ GEN
 RgXQX_pseudorem(GEN x, GEN y, GEN T)
 {
   long vx = varn(x), dx, dy, dz, i, lx, p;
-  pari_sp av = avma, av2, lim;
+  pari_sp av = avma, av2;
   GEN y_lead;
 
   if (!signe(y)) pari_err_INV("RgXQX_pseudorem",y);
@@ -1764,7 +1869,7 @@ RgXQX_pseudorem(GEN x, GEN y, GEN T)
   for (i = 1; i <= dy; i++)
     if (isexactzero(gel(y,i))) gel(y,i) = NULL;
   dz = dx-dy; p = dz+1;
-  av2 = avma; lim = stack_lim(av2,1);
+  av2 = avma;
   for (;;)
   {
     gel(x,0) = gneg(gel(x,0)); p--;
@@ -1781,7 +1886,7 @@ RgXQX_pseudorem(GEN x, GEN y, GEN T)
     }
     do { x++; dx--; } while (dx >= 0 && gequal0(gel(x,0)));
     if (dx < dy) break;
-    if (low_stack(lim,stack_lim(av2,1)))
+    if (gc_needed(av2,1))
     {
       if(DEBUGMEM>1) pari_warn(warnmem,"RgX_pseudorem dx = %ld >= %ld",dx,dy);
       gerepilecoeffs(av2,x,dx+1);
@@ -1817,7 +1922,7 @@ GEN
 RgXQX_pseudodivrem(GEN x, GEN y, GEN T, GEN *ptr)
 {
   long vx = varn(x), dx, dy, dz, i, iz, lx, lz, p;
-  pari_sp av = avma, av2, lim;
+  pari_sp av = avma, av2;
   GEN z, r, ypow, y_lead;
 
   if (!signe(y)) pari_err_INV("RgXQX_pseudodivrem",y);
@@ -1852,7 +1957,7 @@ RgXQX_pseudodivrem(GEN x, GEN y, GEN T, GEN *ptr)
     GEN c = gmul(gel(ypow,i-1), y_lead);
     gel(ypow,i) = rem(c,T);
   }
-  av2 = avma; lim = stack_lim(av2,1);
+  av2 = avma;
   for (iz=2;;)
   {
     p--;
@@ -1871,7 +1976,7 @@ RgXQX_pseudodivrem(GEN x, GEN y, GEN T, GEN *ptr)
     x++; dx--;
     while (dx >= dy && gequal0(gel(x,0))) { x++; dx--; iz++; }
     if (dx < dy) break;
-    if (low_stack(lim,stack_lim(av2,1)))
+    if (gc_needed(av2,1))
     {
       GEN X = x-2;
       if(DEBUGMEM>1) pari_warn(warnmem,"RgX_pseudodivrem dx=%ld >= %ld",dx,dy);
@@ -1929,6 +2034,12 @@ RgXQX_RgXQ_mul(GEN x, GEN y, GEN T)
 {
   return RgXQX_red(RgX_Rg_mul(x,y), T);
 }
+GEN
+RgXQV_RgXQ_mul(GEN v, GEN x, GEN T)
+{
+  return RgXQV_red(RgV_Rg_mul(v,x), T);
+}
+
 GEN
 RgXQX_sqr(GEN x, GEN T)
 {
@@ -1938,6 +2049,8 @@ RgXQX_sqr(GEN x, GEN T)
 static GEN
 _add(void *data, GEN x, GEN y) { (void)data; return RgX_add(x, y); }
 static GEN
+_sub(void *data, GEN x, GEN y) { (void)data; return RgX_sub(x, y); }
+static GEN
 _sqr(void *data, GEN x) { return RgXQ_sqr(x, (GEN)data); }
 static GEN
 _mul(void *data, GEN x, GEN y) { return RgXQ_mul(x,y, (GEN)data); }
@@ -1950,7 +2063,8 @@ _zero(void *data) { return pol_0(varn((GEN)data)); }
 static GEN
 _red(void *data, GEN x) { (void)data; return gcopy(x); }
 
-static struct bb_algebra RgXQ_algebra = { _red,_add,_mul,_sqr,_one,_zero };
+static struct bb_algebra RgXQ_algebra = { _red, _add, _sub,
+              _mul, _sqr, _one, _zero };
 
 GEN
 RgX_RgXQV_eval(GEN Q, GEN x, GEN T)
@@ -1961,7 +2075,7 @@ RgX_RgXQV_eval(GEN Q, GEN x, GEN T)
 GEN
 RgX_RgXQ_eval(GEN Q, GEN x, GEN T)
 {
-  int use_sqr = (degpol(x)<<1) >= degpol(T);
+  int use_sqr = 2*degpol(x) >= degpol(T);
   return gen_bkeval(Q,degpol(Q),x,use_sqr,(void*)T,&RgXQ_algebra,_cmul);
 }
 
@@ -1973,12 +2087,12 @@ struct modXn {
 static GEN
 _sqrXn(void *data, GEN x) {
   struct modXn *S = (struct modXn*)data;
-  return RgX_sqrlow(x, S->n);
+  return RgXn_sqr(x, S->n);
 }
 static GEN
 _mulXn(void *data, GEN x, GEN y) {
   struct modXn *S = (struct modXn*)data;
-  return RgX_mullow(x,y, S->n);
+  return RgXn_mul(x,y, S->n);
 }
 static GEN
 _oneXn(void *data) {
@@ -1990,11 +2104,54 @@ _zeroXn(void *data) {
   struct modXn *S = (struct modXn*)data;
   return pol_0(S->v);
 }
-static struct bb_algebra RgX_modXn_algebra = { _red,_add, _mulXn,_sqrXn, _oneXn,_zeroXn };
+static struct bb_algebra RgXn_algebra = { _red, _add, _sub, _mulXn, _sqrXn,
+                                          _oneXn, _zeroXn };
+
+GEN
+RgXn_powers(GEN x, long m, long n)
+{
+  long d = degpol(x);
+  int use_sqr = (d<<1) >= n;
+  struct modXn S;
+  S.v = varn(x); S.n = n;
+  return gen_powers(x,m,use_sqr,(void*)&S,_sqrXn,_mulXn,_oneXn);
+}
+
+GEN
+RgXn_powu_i(GEN x, ulong m, long n)
+{
+  struct modXn S;
+  S.v = varn(x); S.n = n;
+  return gen_powu_i(x, m, (void*)&S,_sqrXn,_mulXn);
+}
+GEN
+RgXn_powu(GEN x, ulong m, long n)
+{
+  struct modXn S;
+  S.v = varn(x); S.n = n;
+  return gen_powu(x, m, (void*)&S,_sqrXn,_mulXn);
+}
+
+GEN
+RgX_RgXnV_eval(GEN Q, GEN x, long n)
+{
+  struct modXn S;
+  S.v = varn(gel(x,2)); S.n = n;
+  return gen_bkeval_powers(Q,degpol(Q),x,(void*)&S,&RgXn_algebra,_cmul);
+}
+
+GEN
+RgX_RgXn_eval(GEN Q, GEN x, long n)
+{
+  int use_sqr = 2*degpol(x) >= n;
+  struct modXn S;
+  S.v = varn(x); S.n = n;
+  return gen_bkeval(Q,degpol(Q),x,use_sqr,(void*)&S,&RgXn_algebra,_cmul);
+}
 
 /* Q(x) mod t^n, x in R[t], n >= 1 */
 GEN
-RgX_modXn_eval(GEN Q, GEN x, long n)
+RgXn_eval(GEN Q, GEN x, long n)
 {
   long d = degpol(x);
   int use_sqr;
@@ -2007,7 +2164,111 @@ RgX_modXn_eval(GEN Q, GEN x, long n)
   S.v = varn(x);
   S.n = n;
   use_sqr = (d<<1) >= n;
-  return gen_bkeval(Q,degpol(Q),x,use_sqr,(void*)&S,&RgX_modXn_algebra,_cmul);
+  return gen_bkeval(Q,degpol(Q),x,use_sqr,(void*)&S,&RgXn_algebra,_cmul);
+}
+
+GEN
+RgXn_inv(GEN f, long e)
+{
+  pari_sp av = avma, av2;
+  ulong mask;
+  GEN W;
+  long v = varn(f), n=1;
+  if (signe(f)==0)
+    pari_err_INV("RgXn_inv",f);
+  W = scalarpol(ginv(gel(f,2)),v);
+  mask = quadratic_prec_mask(e);
+  av2 = avma;
+  for (;mask>1;)
+  {
+    GEN u, fr;
+    long n2 = n;
+    n<<=1; if (mask & 1) n--;
+    mask >>= 1;
+    fr = RgXn_red_shallow(f, n);
+    u = RgX_shift_shallow(RgXn_mul(W, fr, n), -n2);
+    W = RgX_sub(W, RgX_shift_shallow(RgXn_mul(u, W, n-n2), n2));
+    if (gc_needed(av2,2))
+    {
+      if(DEBUGMEM>1) pari_warn(warnmem,"RgXn_inv, e = %ld", n);
+      W = gerepileupto(av2, W);
+    }
+  }
+  return gerepileupto(av, W);
+}
+
+GEN
+RgXn_exp(GEN h, long e)
+{
+  pari_sp av = avma, av2;
+  long v = varn(h), n=1;
+  GEN f = pol_1(v), g = pol_1(v);
+  ulong mask = quadratic_prec_mask(e);
+  av2 = avma;
+  if (signe(h)==0 || degpol(h)<1 || !gequal0(gel(h,2)))
+    pari_err_DOMAIN("RgXn_exp","valuation", "<", gen_1, h);
+  for (;mask>1;)
+  {
+    GEN q, w;
+    long n2 = n;
+    n<<=1; if (mask & 1) n--;
+    mask >>= 1;
+    g = RgX_sub(RgX_muls(g,2),RgXn_mul(f,RgXn_sqr(g,n2),n2));
+    q = RgX_deriv(RgXn_red_shallow(h,n2));
+    w = RgX_add(q, RgXn_mul(g, RgX_sub(RgX_deriv(f), RgXn_mul(f,q,n-1)),n-1));
+    f = RgX_add(f, RgXn_mul(f, RgX_sub(RgXn_red_shallow(h, n), RgX_integ(w)), n));
+    if (gc_needed(av2,2))
+    {
+      if(DEBUGMEM>1) pari_warn(warnmem,"RgXn_exp, e = %ld", n);
+      gerepileall(av2, 2, &f, &g);
+    }
+  }
+  return gerepileupto(av, f);
+}
+
+GEN
+RgXn_reverse(GEN f, long e)
+{
+  pari_sp av = avma, av2;
+  ulong mask;
+  GEN fi, a, df, W, an;
+  long v = varn(f), n=1;
+  if (degpol(f)<1 || !gequal0(gel(f,2)))
+    pari_err_INV("serreverse",f);
+  fi = ginv(gel(f,3));
+  a = deg1pol_shallow(fi,gen_0,v);
+  if (e <= 2) return gerepilecopy(av, a);
+  W = scalarpol(fi,v);
+  df = RgX_deriv(f);
+  mask = quadratic_prec_mask(e);
+  av2 = avma;
+  for (;mask>1;)
+  {
+    GEN u, fa, fr;
+    long n2 = n, rt;
+    n<<=1; if (mask & 1) n--;
+    mask >>= 1;
+    fr = RgXn_red_shallow(f, n);
+    rt = brent_kung_optpow(degpol(fr), 4, 3);
+    an = RgXn_powers(a, rt, n);
+    if (n>1)
+    {
+      long n4 = (n2+1)>>1;
+      GEN dfr = RgXn_red_shallow(df, n2);
+      dfr = RgX_RgXnV_eval(dfr, RgXnV_red_shallow(an, n2), n2);
+      u = RgX_shift(RgX_Rg_sub(RgXn_mul(W, dfr, n2), gen_1), -n4);
+      W = RgX_sub(W, RgX_shift(RgXn_mul(u, W, n2-n4), n4));
+    }
+    fa = RgX_sub(RgX_RgXnV_eval(fr, an, n), pol_x(v));
+    fa = RgX_shift(fa, -n2);
+    a = RgX_sub(a, RgX_shift(RgXn_mul(W, fa, n-n2), n2));
+    if (gc_needed(av2,2))
+    {
+      if(DEBUGMEM>1) pari_warn(warnmem,"RgXn_reverse, e = %ld", n);
+      gerepileall(av2, 2, &a, &W);
+    }
+  }
+  return gerepileupto(av, a);
 }
 
 /* x,T in Rg[X], n in N, compute lift(x^n mod T)) */
@@ -2044,7 +2305,7 @@ RgXQ_pow(GEN x, GEN n, GEN T)
 GEN
 RgXQ_powers(GEN x, long l, GEN T)
 {
-  int use_sqr = (degpol(x)<<1) >= degpol(T);
+  int use_sqr = 2*degpol(x) >= degpol(T);
   return gen_powers(x, l, use_sqr, (void *)T,_sqr,_mul,_one);
 }
 
@@ -2070,9 +2331,15 @@ QXQ_powers(GEN a, long n, GEN T)
 static GEN
 do_QXQ_eval(GEN v, long imin, GEN a, GEN T)
 {
-  long l, i, m = degpol(T);
-  GEN dz, z = Q_remove_denom(QXQ_powers(a, m-1, T), &dz);
+  long l, i, m = 0;
+  GEN dz, z;
   GEN V = cgetg_copy(v, &l);
+  for (i = imin; i < l; i++)
+  {
+    GEN c = gel(v, i);
+    if (typ(c) == t_POL) m = maxss(m, degpol(c));
+  }
+  z = Q_remove_denom(QXQ_powers(a, m, T), &dz);
   for (i = 1; i < imin; i++) V[i] = v[i];
   for (i = imin; i < l; i++)
   {
@@ -2114,7 +2381,7 @@ RgXQ_norm(GEN x, GEN T)
   GEN L, y;
 
   av = avma; y = resultant(T, x);
-  L = leading_term(T);
+  L = leading_coeff(T);
   if (gequal1(L) || !signe(x)) return y;
   return gerepileupto(av, gdiv(y, gpowgs(L, dx)));
 }
